# Copyright (C) 2020 Ludvig Larsson aka Valmond

import os
import sys
from Crypto.PublicKey import RSA
from Crypto.Cipher import AES
from Crypto.Util import Counter
import binascii
import json
import tarfile
import sqlite3
from sqlite3 import Error

from socket_communication import SecureCommunication
import database_query as dq
import database_connection as dc


# This is the client software that takes a link.10f and gets the according data, decrypts it and saves it to disc

# Link file: DEPRECATED, all this data is now stored in a SQLite database file
#
# IP
# PORT
# NAME (use this one to get the encrypted file)
# ORIGFILENAME (use this one as a filename for the decrypted file)
# AES (64 hex chars)
#

# Example (with the rsa_pub keys shortened for readability
'''
{"name": "music",
"originalfilename": "Marina Topley Bird Track 01.mp3",
"key": "c389edc914cd375c5367ca50bcc80591f6ef11b18db3ae3c850628d6f9cf5f66", "iv": "0313bc0a41820832406fb28980f4cadd",
"data_rsa_pub": "30820122300d060",

"address_list": 
[ {"rsa_pub": "122300d0609", "ip": "127.0.0.1", "port": 12000}
, {"rsa_pub": "4886f70d011", "ip": "127.0.0.1", "port": "11000"}
, {"rsa_pub": "3006092a864", "ip": "127.0.0.1", "port": "10000"}
]

}
'''


def print_help():
    print('Usage:')
    print('')
    print('python3 get_data_from_link.py link.10f')
    print('python3 get_data_from_link.py link.10f dest_folder')
    print('This both works for files and folders')
    # print('python3 get_data_from_link.py --testrun')

    sys.exit()

# AES supports multiple key sizes: 16 (AES128), 24 (AES192), or 32 (AES256).
key_bytes = 32

# Takes as input a 32-byte key, a 16-byte IV, and a ciphertext, and outputs the
# corresponding plaintext.
def decrypt(key, iv, ciphertext):
    assert len(key) == key_bytes

    # Initialize counter for decryption. iv should be the same as the output of
    # encrypt().
    # iv_int = int(iv.encode('hex'), 16)
    iv_int = int(binascii.hexlify(iv), 16) 
    ctr = Counter.new(AES.block_size * 8, initial_value=iv_int)

    # Create AES-CTR cipher.
    aes = AES.new(key, AES.MODE_CTR, counter=ctr)

    # Decrypt and return the plaintext.
    plaintext = aes.decrypt(ciphertext)
    return plaintext


# If there is a database, lets store the addresses there so that we will know more nodes when swicharooing and so
def save_off_addresses(address_list):
    if dc.database_exists():
        for address in address_list:
            if 'rsa_pub' in address and 'ip' in address and 'port' in address:
                rsa_hex_pub_key = address['rsa_pub']
                ip = address['ip']
                port = address['port']

                # Insert nodes from the link, if the database exists so we can download data without a database
                dq.insert_node(ip, port, rsa_hex_pub_key)


# Gets all the data from the link and returns it, or None if failed
# info, which contains: name, originalfilename, key, iv, data_rsa_pub
# address_list, which is a list of addesses:
def extract_data_from_database_link(database_file_name):
    try:
        connection = sqlite3.connect(database_file_name)
    except Error:
        print("Could not create the (database) link:")
        print(Error)
        return None

    cursor = connection.cursor()
    cursor.execute("SELECT name, originalfilename, key, iv, data_rsa_pub FROM information")
    rows = cursor.fetchall()
    if len(rows) == 0:
        print("Error: No information in link")
        return None
    data = rows[0]
    info = {
        'name': data[0],
        'originalfilename': data[1],
        'key': str(data[2]),
        'iv': data[3],
        'data_rsa_pub': data[4]
    }

    cursor.execute("SELECT rsa_pub, ip, port FROM address_list")
    rows = cursor.fetchall()
    if len(rows) == 0:
        print("Error: There are no addresses in the link")
        return None
    address_list = []
    for data in rows:  # maybe this is not necessary... but as it should work, lets check that later
        address = {
            'rsa_pub': data[0],
            'ip': data[1],
            'port': data[2]
        }
        address_list.append(address)

    return info, address_list



def get_data_from_database_link(linkfile, dest_folder):
    # use bytes.fromhex()
    print('get link')

    linkfile = linkfile.replace('\\', '/')
    dest_folder = dest_folder.replace('\\', '/')

    info, address_list = extract_data_from_database_link(linkfile)

    print('Download :', info['name'], ' original name:', info['originalfilename'])

    for address in address_list:
        print('Address:', address['ip'])


    # this is the key,iv for the encrypted file we want to retrieve
    print('TYPE of key:', type(info['key']))
    print('VAL  of key:', info['key'])
    key = bytes.fromhex(info['key'])
    iv = bytes.fromhex(info['iv'])

    data_rsa_pub = info['data_rsa_pub']

    tries = 3
    timeout_seconds = 0.3  # We'll add to this if no address has the time to answer

    while tries > 0:

        # Loop all adresses and try to get the data. todo: make this parallel / or at least bail out faster if node is not up.
        for address in address_list:
            if 'rsa_pub' in address and 'ip' in address and 'port' in address:
                rsa_hex_pub_key = address['rsa_pub']
                ip = address['ip']
                port = address['port']

                rsa_pub_key = RSA.importKey(bytes.fromhex(rsa_hex_pub_key))
            else:
                print('#Error: address corrupt.')

            print('set up secure link:', ip, port)
            sys.stdout.flush()
            sc = SecureCommunication(ip, port, rsa_pub_key,
                                     timeout=timeout_seconds)  # rsa_pub here is the one linked to the DATA, not the node we are connecting to!
            if sc.is_active():
                print('query for data')
                sys.stdout.flush()
                query = 'ONEFINGER:' + info['name'] + ',' + data_rsa_pub  # rsa in plaintext here as we use it as a name
                encrypted_data = sc.send(query)
                if encrypted_data is not None:
                    print('Decrypt data')
                    plain_data = decrypt(key, iv, encrypted_data)
                    originalfilename = info['originalfilename']
                    # Should we really unzip ALL zipped files??
                    is_gzipped = originalfilename[-7:] == '.tar.gz'
                    print('File extension check:', originalfilename, '=>', originalfilename[-7:], ' result=', is_gzipped)
                    if dest_folder:
                        if not os.path.exists(dest_folder):
                            os.makedirs(dest_folder)
                        originalfilename = os.path.join(dest_folder, originalfilename)
                    with open(originalfilename, 'wb') as f:
                        f.write(plain_data)
                        f.close()
                        print('Success: retrieved data.')

                        # was it a folder? If so, decompress it here!
                        if is_gzipped:
                            print('Untar and gunzip')
                            try:
                                with tarfile.open(originalfilename, "r:gz") as tar_handle:
                                    dir = dest_folder if dest_folder else './'
                                    tar_handle.extractall(path=dir)  # Will be made here in this folder
                                    tar_handle.close()
                            except BaseException as error:
                                print('An exception occurred: {}'.format(error))

                        return
                else:
                    print('SecureCommunication closed down too early')
            else:
                print('SecureCommunication Failed')
        tries -= 1
        timeout_seconds = (timeout_seconds + 1) * 2  # So if starting at 0.3,:  0.3, 2.6., 7.2


def get_data(linkfile, dest_folder):
    # use bytes.fromhex()
    print('get link')

    linkfile = linkfile.replace('\\', '/')
    dest_folder = dest_folder.replace('\\', '/')

    # New secure link
    # both name and secure encryption
    with open(linkfile) as json_data:
        print('get json')
        d = json.load(json_data)
        # save off addresses for later use in swicharoo etc:
        if 'address_list' in d:
            save_off_addresses(d['address_list'])

        # Start the actual downloading
        if 'address_list' in d and 'key' in d and 'iv' in d and 'data_rsa_pub' in d:
            address_list = d['address_list']

            # this is the key,iv for the encrypted file we want to retrieve
            key = bytes.fromhex(d['key'])
            iv = bytes.fromhex(d['iv'])
            data_rsa_pub = d['data_rsa_pub']

            tries = 3
            timeout_seconds = 0.3  # We'll add to this if no address has the time to answer

            while tries > 0:

                # Loop all adresses and try to get the data. todo: make this parallel / or at least bail out faster if node is not up.
                for address in address_list:
                    if 'rsa_pub' in address and 'ip' in address and 'port' in address:
                        rsa_hex_pub_key = address['rsa_pub']
                        ip = address['ip']
                        port = address['port']

                        rsa_pub_key = RSA.importKey(bytes.fromhex(rsa_hex_pub_key))
                    else:
                        print('#Error: address corrupt.')

                    print('set up secure link:', ip, port)
                    sys.stdout.flush()
                    sc = SecureCommunication(ip, port, rsa_pub_key, timeout=timeout_seconds)  # rsa_pub here is the one linked to the DATA, not the node we are connecting to!
                    if sc.is_active():
                        print('query for data')
                        sys.stdout.flush()
                        query = 'ONEFINGER:' + d['name'] + ',' + data_rsa_pub  # rsa in plaintext here as we use it as a name
                        encrypted_data = sc.send(query)
                        if encrypted_data is not None:
                            print('Decrypt data')
                            plain_data = decrypt(key, iv, encrypted_data)
                            originalfilename = d['originalfilename']
                            # Should we really unzip ALL zipped files??
                            is_gzipped = originalfilename[-7:] == '.tar.gz'
                            print('File extension check:', originalfilename, '=>', originalfilename[-7:], ' result=', is_gzipped)
                            if dest_folder:
                                if not os.path.exists(dest_folder):
                                    os.makedirs(dest_folder)
                                originalfilename = os.path.join(dest_folder, originalfilename)
                            with open(originalfilename, 'wb') as f:
                                f.write(plain_data)
                                f.close()
                                print('Success: retrieved data.')

                                # was it a folder? If so, decompress it here!
                                if is_gzipped:
                                    print('Untar and gunzip')
                                    try:
                                        with tarfile.open(originalfilename, "r:gz") as tar_handle:
                                            dir = dest_folder if dest_folder else './'
                                            tar_handle.extractall(path=dir)  # Will be made here in this folder
                                            tar_handle.close()
                                    except BaseException as error:
                                        print('An exception occurred: {}'.format(error))

                                return
                        else:
                            print('SecureCommunication closed down too early')
                    else:
                        print('SecureCommunication Failed')
                tries -= 1
                timeout_seconds = (timeout_seconds + 1) * 2  # So if starting at 0.3,:  0.3, 2.6., 7.2


        else:
            print('#Error: Missing information in link file.')


# Test case:
def test():
    print('[get_data_from_link.py.py]')
    print('Test the get_data_from_link:')
    get_data('test.10f')
    get_data('image.10f')




if __name__ == '__main__':
    if len(sys.argv) == 1:
        print_help()

    if sys.argv[1] == '--testrun':
        print('--testrun')
        test()
    else:
        print('manually get data')
        dest_folder = None
        if len(sys.argv) >= 3:
            dest_folder = sys.argv[2]
        get_data_from_database_link(sys.argv[1], dest_folder)

